package org.privas.data;

import com.opencsv.CSVParser;
import com.opencsv.CSVParserBuilder;
import com.opencsv.CSVReader;
import com.opencsv.CSVReaderBuilder;
import org.privas.parser.AttributesOptions;
import org.privas.parser.PrivacyDefinition;
import org.privas.parser.TableToPrivateDetails;

import java.io.IOException;
import java.io.Reader;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.sql.Types;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

public class CSVExtractor implements Extractor<TableData> {

    private String path;
    private Reader reader;

    public CSVExtractor(PrivacyDefinition privacyDefinition) {
        this.path = privacyDefinition.getPath();
    }

    @Override
    public Collection<TableData> extract(PrivacyDefinition privacyDefinition) {
        List<TableData> data = new ArrayList<>();

        // for each table, one in CSV case:
        for (TableToPrivateDetails tableDetail : privacyDefinition.getTableToPrivateDetails()) {
            // read table
            TableData tableData = getTable(tableDetail.getTableName(), tableDetail.getAttributesOptions(), privacyDefinition);

            data.add(tableData);
        }

        return data;
    }

    private TableData getTable(String name, AttributesOptions attributesOptions, PrivacyDefinition privacyDefinition) {
        TableData tableData = null;

        try {
            List<String[]> list = new ArrayList<>();
            CSVReader csvReader = getCSVReader();
            String[] line;

            tableData = new TableData(name, attributesOptions, privacyDefinition.getGeneralizations());

            String[] headers = csvReader.readNext();

            for (int i = 0; i < headers.length; i++) {
                String header = headers[i];

                tableData.addColumnName(header, "VARCHAR", Types.VARCHAR, 1024, i - 1, true);
            }

            tableData.consolidateColumns();

            while ((line = csvReader.readNext()) != null) {
                Row row = new Row();
                boolean validLine = true;

                for (int i = 0; i < line.length && validLine; i++) {
                    String value = line[i];

                    if (value == null || value.trim().equals("?")) {
                        validLine = false;
                    }
                    else {
                        row.addValue(value.trim());
                    }
                }

                if (line.length > 0 && validLine) {
                    tableData.add("", row);
                }
            }

            close(csvReader);

            return tableData;
        } catch (IOException e) {
            e.printStackTrace();
        }

        return tableData;
    }

    private CSVReader getCSVReader() {
        try {
            Path logFile = Paths.get(path);

            reader = Files.newBufferedReader(logFile);

            CSVParser parser = new CSVParserBuilder()
                    .withSeparator(',')
                    .withIgnoreQuotations(true)
                    .build();

            CSVReader csvReader = new CSVReaderBuilder(reader)
                    .withSkipLines(0)
                    .withCSVParser(parser)
                    .build();

            return csvReader;

        } catch (IOException e) {
            e.printStackTrace();
        }

        return null;
    }

    private void close(CSVReader csvReader) {
        try {
            reader.close();
            csvReader.close();
        } catch (IOException e) {
            e.printStackTrace();
        }
    }
}
