package org.privas.data;

import org.privas.parser.Attribute;
import org.privas.parser.GeneralizationRange;
import org.privas.parser.GeneralizationString;

import java.sql.Types;
import java.time.LocalDateTime;
import java.util.Collection;
import java.util.List;
import java.util.stream.Collectors;

public class ColumnMetaData {
    // column name
    private String name;

    // sql type (varchar, integer, etc)
    private String dataTypeLabel;

    private int dataType;

    // column length
    private int length;

    private int currentPosition;

    private int originalPosition;

    private boolean nullable;

    private boolean isQId;

    private boolean isSA;

    private boolean isRangeType;

    private Attribute attribute;

    public ColumnMetaData(String name, String dataTypeLabel, int dataType, int length, int originalPosition, boolean nullable, boolean isQId, boolean isSA, Attribute attribute) {
        this.name = name;
        this.dataTypeLabel = dataTypeLabel;
        this.dataType = dataType;
        this.length = length;
        this.originalPosition = originalPosition;
        this.nullable = nullable;
        this.isQId = isQId;
        this.isSA = isSA;
        this.isRangeType = !attribute.hasType();
        this.attribute = attribute;
    }

    public void setCurrentPosition(int currentPosition) {
        this.currentPosition = currentPosition;
    }

    @Override
    public String toString(){
        dataTypeLabel = dataTypeLabel.replace(" UNSIGNED", "");
        // SMALLINT UNSIGNED
        // SMALLINT(5) UNSIGNED
        return name + " " + dataTypeToSQLType();
    }

    private String dataTypeToSQLType() {
        String nullableStr = nullable ? " NULL" : "";

        switch (dataType) {
            case Types.ARRAY:
                return "VARCHAR (" + 1024 + ")" + nullableStr;
            case Types.SMALLINT:
                return "VARCHAR (" + 1024 + ")" + nullableStr;
            case Types.VARCHAR:
                if (length <= 40) {
                    return "VARCHAR (" + 1024 + ")" + nullableStr;
                }
                return "VARCHAR (" + 1024 + ")" + nullableStr;

            case Types.TIMESTAMP:
                return "VARCHAR (" + 1024 + ")" + nullableStr;
                //return "TIMESTAMP";
            case Types.DECIMAL:
                return "VARCHAR (" + 1024 + ")" + nullableStr;
            case Types.TINYINT:
                return "VARCHAR (" + 1024 + ")" + nullableStr;
            case Types.BIT:
                return "VARCHAR (" + 1024 + ")" + nullableStr;
            case Types.INTEGER:
                return "VARCHAR (" + 1024 + ")" + nullableStr;
            case Types.LONGVARBINARY:
                return "BLOB" + nullableStr;
            case Types.BLOB:
                return "BLOB" + nullableStr;
            default:
                return "VARCHAR (" + 1024 + ")" + nullableStr;
        }
    }

    private String dataTypeToSQLTypeOriginal() {
        switch (dataType) {
            case Types.ARRAY:
                return "ARRAY";
            case Types.SMALLINT:
                return "SMALLINT";
            case Types.VARCHAR:
                if (length <= 40) {
                    return "VARCHAR (" + 1024 + ")";
                }
                return "VARCHAR (" + length + ")";

            case Types.TIMESTAMP:
                return "VARCHAR (" + 1024 + ")";
            //return "TIMESTAMP";
            case Types.DECIMAL:
                return "DECIMAL (" + length + ")";
            case Types.TINYINT:
                return "TINYINT (" + length + ")";
            case Types.BIT:
                return "TINYINT (" + length + ")";
            case Types.INTEGER:
                return "INT (" + length + ")";
            case Types.LONGVARBINARY:
                return "BLOB";
            case Types.BLOB:
                return "BLOB";
            default:
                return "";
        }
    }

    public String getName() {
        return name;
    }

    public String getDataTypeLabel() {
        return dataTypeLabel;
    }

    public int getDataType() {
        return dataType;
    }

    public int getLength() {
        return length;
    }

    public int getCurrentPosition() {
        return currentPosition;
    }

    public int getOriginalPosition() {
        return originalPosition;
    }

    public boolean isQId() {
        return isQId;
    }

    public boolean isSA() {
        return isSA;
    }

    public boolean isToPrivate() {
        return isQId() || isSA();
    }

    public boolean isRangeType() {
        return isRangeType;
    }

    public Attribute getAttribute() {
        return attribute;
    }

    public GeneralizationRange getGeneralizationRange(List<Row> data) {
        if (!isRangeType) {
            return null;
        }

        DataType privasDataType = getPrivasDataType(dataType);

        if (privasDataType == DataType.NUMBER) {
            int[] numbers = data.stream()
                    .map(row -> row.getValue(currentPosition))
                    .mapToInt(Integer::parseInt)
                    .toArray();

            return new GeneralizationRange(numbers);
        }
        else if (privasDataType == DataType.DATE) {
            LocalDateTime[] dates = (LocalDateTime[]) data.stream()
                    .map(row -> row.getValue(currentPosition))
                    .map(LocalDateTime::parse)
                    .toArray();

            return new GeneralizationRange(dates);
        }
        else {
            Collection<String> values = data.stream()
                    .map(row -> row.getValue(currentPosition))
                    .collect(Collectors.toList());

            return new GeneralizationString(values);
        }
    }

    private boolean isRangeTypeFrom(int dataT) {
        DataType dataType = getPrivasDataType(dataT);

        return dataType == DataType.NUMBER || dataType == DataType.DATE;
    }

    private DataType getPrivasDataType(int dataType) {
        switch (dataType) {
            case Types.INTEGER:
            case Types.SMALLINT:
            case Types.DECIMAL:
                return DataType.NUMBER;
            case Types.TIMESTAMP:
                return DataType.DATE;
            default:
                return DataType.OTHER;
        }
    }
}

