package org.privas.parser;

import java.util.ArrayList;
import java.util.List;
import java.util.Optional;

public class GeneralizationNode {
    private GeneralizationNode parent;

    private String label;
    private boolean isLeaf;
    private List<GeneralizationNode> children;

    public GeneralizationNode(GeneralizationNode parent, String label) {
        this(parent, label, false);
    }

    public GeneralizationNode(GeneralizationNode parent, String label, boolean isLeaf) {
        this.parent = parent;
        children = new ArrayList<>();
        init(label, isLeaf);
    }

    public void init(String label, boolean isLeaf) {
        this.label = label;
        this.isLeaf = isLeaf;
    }

    public void add(GeneralizationNode node) {
        if (!isLeaf) {
            this.children.add(node);
        }
    }

    public void add(List<GeneralizationNode> nodes) {
        this.children.addAll(nodes);
    }

    public String getLabel() {
        return label;
    }

    public boolean isLeaf() {
        return isLeaf;
    }

    public GeneralizationNode getParent() {
        return parent;
    }

    public void setParent(GeneralizationNode parent) {
        this.parent = parent;
    }

    public List<GeneralizationNode> getChildren() {
        return children;
    }

    public List<GeneralizationNode> getAllNodes() {
        List<GeneralizationNode> nodes = new ArrayList<>(children);

        for (GeneralizationNode child: children) {
            nodes.addAll(child.getAllNodes());
        }

        return nodes;
    }

    public int getNumberOfLeafsCovered() {
        if (isLeaf) {
            return 1;
        }

        return children.stream()
                .mapToInt(GeneralizationNode::getNumberOfLeafsCovered)
                .sum();
    }

    public Optional<GeneralizationNode> getGeneralizationFor(String value) {
        if (label.equals(value)) {
            return Optional.of(this);
        }

        return children.stream()
                .map(child -> child.getGeneralizationFor(value))
                .filter(Optional::isPresent)
                .map(Optional::get)
                .findFirst();
    }

    public boolean covers(String value) {
        if (label.equals(value)) {
            return true;
        }

        return children.stream()
                .anyMatch(child -> child.covers(value));
    }

    public int height() {
        if (isLeaf) {
            return 0;
        }

        return 1 + children.stream().mapToInt(GeneralizationNode::height).max().orElse(1);
    }
}
