package org.privas.runtime;

import org.antlr.v4.runtime.ANTLRInputStream;
import org.antlr.v4.runtime.CommonTokenStream;
import org.privas.parser.PrivacyDefinition;
import org.privas.PrivasLexer;
import org.privas.PrivasParser;
import org.privas.data.DataConnectorFactory;
import org.privas.data.Exporter;
import org.privas.data.Extractor;
import org.privas.data.TableDataConnectorFactory;
import org.privas.data.TableData;
import org.privas.techniques.Model;

import java.io.FileInputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

public class PrivasMain {

    private static DecisionEngine decisionEngine = new DecisionEngine();
    private static DataConnectorFactory<TableData> dataDataConnectorFactory = new TableDataConnectorFactory();

    public static void main(String[] args) {
        String filePath = extractPath(args);

        if (filePath == null) {
            return;
        }

        PrivacyDefinition privacyDefinition = parse(filePath);

        if (hasErrors(privacyDefinition)) {
            handleErrors(privacyDefinition, filePath);

            return;
        }

        Extractor<TableData> dataExtractor = dataDataConnectorFactory.getExtractor(privacyDefinition);

        Collection<TableData> data = dataExtractor.extract(privacyDefinition);
        List<TableData> dataTransformed = new ArrayList<>();

        for (TableData tableData: data) {
            Model<TableData> privacyModel = decisionEngine.getModel(privacyDefinition);

            privacyModel.addData(tableData);
            privacyModel.apply();
            dataTransformed.add(privacyModel.getDataTransformed());
        }

        Exporter<TableData> dataExporter = dataDataConnectorFactory.getExporter(privacyDefinition);

        dataExporter.init();

        for (TableData tableData: dataTransformed) {
            dataExporter.export(tableData);
        }
    }

    private static PrivacyDefinition parse(String filePath) {
        try {
            ANTLRInputStream antlrInputStream = new ANTLRInputStream(new FileInputStream(filePath));

            PrivasLexer lexer = new PrivasLexer(antlrInputStream);

            CommonTokenStream tokens = new CommonTokenStream( lexer );
            PrivasParser parser = new PrivasParser( tokens );

            return parser.privas().privacyDefinition;
        }
        catch (IOException e)
        {
            // error
            return null;
        }
    }

    private static boolean hasErrors(PrivacyDefinition privacyDefinition) {
        return privacyDefinition == null || privacyDefinition.hasErrors();
    }

    private static void handleErrors(PrivacyDefinition privacyDefinition, String filePath) {
        if (privacyDefinition == null) {
            System.out.println();
            System.out.println("Error :: Not able to parse the PrivasL file `"+ filePath +"`");

            return;
        }
        System.out.println();
        System.out.println("Error :: The PrivasL file has errors. The following errors were found:");

        privacyDefinition.getErrors().stream().forEach(error -> System.out.println("\t- " + error));
        System.out.println();
    }

    private static String extractPath(String[] args) {
        if (args.length == 0) {
            System.out.println();
            System.out.println("Error :: You should provide the path to the PrivasL definition file.");
            System.out.println("e.g. `java privas.jar FILE.privas`");
            System.out.println();

            return null;
        }

        return args[0];
    }
}
