package org.privas.techniques;

import org.privas.data.Row;
import org.privas.data.TableData;
import org.privas.parser.Generalization;
import org.privas.parser.GeneralizationNode;
import org.privas.parser.GeneralizationRange;

import java.util.*;

import static java.util.stream.Collectors.toMap;

public class Mondrian implements Technique<TableData> {
    private int k;

    private int numberOfQIds = 10;
    private List<Partition> partitionsFinished = new ArrayList<>();
    private List<Generalization> generalizations = new ArrayList<>();
    private List<Integer> QI_RANGE = new ArrayList<>();
    private List<Boolean> isGeneralizationCategorical = new ArrayList<>();

    private TableData tableData;
    private int nrTotalElements = 0;
    private List<List<String>> data = new ArrayList<>();
    private List<List<String>> dataTransformed = new ArrayList<>();

    public Mondrian(int k) {
        this.k = k;
    }

    public double get_normalized_width(Partition partition, int index) {
        /*
        return Normalized width of partition
        similar to NCP
        */
        double width = 0;

        if (!isGeneralizationCategorical.get(index)) {
            Tuple<Integer, Integer> tuple = partition.getWidth().get(index);

            int low = tuple.getFirst();
            int high = tuple.getSecond();

            GeneralizationRange element = (GeneralizationRange) generalizations.get(index);

            if (element.isRange()) {
                width = element.getValuePosition(high) - element.getValuePosition(low);
            }
        } else {
            width = (int) partition.getWidth().get(index).getFirst();
        }

        return width * 1.0 / QI_RANGE.get(index);
    }

    public int chooseQId(Partition partition) {
        /*choose QId with largest normalized Width
        return QId index.
        */
        double maxWitdh = -1;
        int maxQId = -1;

        for (int i = 0; i < numberOfQIds; i++) {
            if (partition.isAllowed(i)) {
                double normWidth = get_normalized_width(partition, i);

                if (normWidth > maxWitdh) {
                    maxWitdh = normWidth;
                    maxQId = i;
                }
            }
        }

        if (maxWitdh > 1) {
            System.out.println("Error: maxWitdh > 1");
            throw new RuntimeException("Error: maxWitdh > 1");
        }

        if (maxQId == -1) {
            System.out.println("cannot find the max QId");
            throw new RuntimeException("cannot find the max QId");
        }

        return maxQId;
    }

    public Map<String, Integer> getFrequencySet(Partition partition, int attributePos) {
        return getFrequencySet(partition.getRecords(), attributePos);
    }

    public Map<String, Integer> getFrequencySet(List<List<String>> records, int attributePos) {
        /*get the frequency_set of partition on dim
        return dict{key: str values, values: count}
        */
        Map<String, Integer> frequency = new TreeMap<>();

        for (List<String> record : records) {
            //element = record[dim]
            String element = record.get(attributePos);
            if (frequency.containsKey(element)) {
                frequency.put(element, frequency.get(element) + 1);
            } else {
                frequency.put(element, 1);
            }
        }

        return frequency;
    }

    public MedianResult find_median(Partition partition, int dim) {
        /*find the middle of the partition
        return splitVal
        */
        Map<String, Integer> frequency = getFrequencySet(partition, dim);
        String split_val = "";
        String next_val = null;
        List<String> value_list = new ArrayList<>(frequency.keySet());

        value_list.sort(new DataComparator());
        int total = frequency.values().stream().mapToInt(a -> a).sum();
        int middle = total / 2;

        if (middle < k || value_list.size() <= 1) {
            if (value_list.isEmpty()) {
                return new MedianResult(null, null, null, null);
            }

            return new MedianResult(null, null, value_list.get(0), value_list.get(value_list.size() - 1));
        }

        int index = 0;

        int split_index = 0;
        boolean alreadyEncountered = false;

        for (int i = 0; i < value_list.size() && !alreadyEncountered; i++) {
            String qi_value = value_list.get(i);

            index += frequency.get(qi_value);

            if (index >= middle) {
                split_val = qi_value;
                split_index = i;
                alreadyEncountered = true;
            }
        }

        if (!alreadyEncountered) {
            System.out.println("Error: cannot find splitVal");
            throw new RuntimeException("Error: cannot find splitVal");
        }

        if (split_index + 1 < value_list.size()) {
            next_val = value_list.get(split_index + 1);
        } else {
            // there is a frequency value in partition
            // which can be handle by mid_set
            // e.g.[1, 2, 3, 4, 4, 4, 4]
            next_val = split_val;
        }

        return new MedianResult(split_val, next_val, value_list.get(0), value_list.get(value_list.size() - 1));
    }


    private Tuple<String, String> split_numerical_value(String numeric_value, Object splitVal) {
    /*
    split numeric value on splitVal
    return sub ranges
    */
        String[] split_num = numeric_value.split(",");
        String lvalue = null;
        String rvalue = null;

        if (split_num.length <= 1) {
            lvalue = split_num[0];
            rvalue = split_num[0];
        } else {
            String low = split_num[0];
            String high = split_num[1];
            // Fix 2,2 problem
            if (low == splitVal) {
                lvalue = low;
            } else {
                lvalue = low + ',' + splitVal;
            }
            if (high == splitVal) {
                rvalue = high;
            } else {
                rvalue = splitVal.toString() + ',' + high;
            }

        }

        return new Tuple<>(lvalue, rvalue);
    }

    private List<Partition> split_numerical(Partition partition, int dim, List<Tuple<Integer, Integer>> pwidth, CurrentGeneralizations currentGeneralizations) {
    /*
        strict split numeric attribute by finding a median,
        lhs = [low, means], rhs = (mean, high]
    */
        List<Partition> subPartitions = new ArrayList<>();
        GeneralizationRange generalization = (GeneralizationRange) generalizations.get(dim);

        // numeric attributes
        MedianResult medianResult = find_median(partition, dim);
        String splitVal = medianResult.getSplitValue();
        String nextVal = medianResult.getNextValue();
        String low = medianResult.getLowValue();
        String high = medianResult.getHighValue();

        Integer lowPosition = generalization.getValuePosition(low);
        Integer highPosition = generalization.getValuePosition(high);

        // update current generalizations
        if (low.equals(high)) {
            currentGeneralizations.setGeneralization(dim, low);
        } else {
            currentGeneralizations.setGeneralization(dim, low + ',' + high);
        }

        // (p_low, p_high)
        pwidth.set(dim, new Tuple<>(lowPosition, highPosition));

        if (splitVal == null || splitVal.equals("") || splitVal.equals(nextVal)) {
            // update current generalizations
            return new ArrayList<>();
        }

        int middlePosition = generalization.getValuePosition(splitVal);
        CurrentGeneralizations leftCurrentGeneralizations = currentGeneralizations.clone();
        CurrentGeneralizations rightCurrentGeneralizations = currentGeneralizations.clone();

        Tuple<String, String> tuple = split_numerical_value(currentGeneralizations.get(dim).toString(), splitVal);
        leftCurrentGeneralizations.setGeneralization(dim, tuple.getFirst());
        rightCurrentGeneralizations.setGeneralization(dim, tuple.getSecond());

        List<List<String>> lhs = new ArrayList<>();
        List<List<String>> rhs = new ArrayList<>();

        for (List<String> temp : partition.getRecords()) {
            String rowElement = temp.get(dim);

            if (generalization.isBeforeOrEqual(rowElement, middlePosition)) {
                // lhs = [low, means]
                lhs.add(temp);
            } else {
                // rhs = (mean, high]
                rhs.add(temp);
            }
        }
        List<Tuple<Integer, Integer>> lwidth = new ArrayList<>(pwidth);
        List<Tuple<Integer, Integer>> rwidth = new ArrayList<>(pwidth);

        lwidth.set(dim, new Tuple<>(pwidth.get(dim).getFirst(), middlePosition));
        rwidth.set(dim, new Tuple<>(generalization.getValuePosition(nextVal), pwidth.get(dim).getSecond()));

        subPartitions.add(new Partition(lhs, lwidth, leftCurrentGeneralizations, numberOfQIds));
        subPartitions.add(new Partition(rhs, rwidth, rightCurrentGeneralizations, numberOfQIds));

        return subPartitions;
    }

    private List<Partition> splitCategorical(Partition partition, int columnQId, List<Tuple<Integer, Integer>> pwidth, CurrentGeneralizations currentGeneralizations) {
        List<Partition> subPartitions = new ArrayList<>();

        String qIDCurrentGeneralization = partition.getCurrentQIdsGeneralization().get(columnQId).getValue();
        GeneralizationNode generalization = generalizations.get(columnQId).getGeneralizationFor(qIDCurrentGeneralization);

        List<GeneralizationNode> subGeneralizationNodes = generalization.getAllNodes();

        List<List<List<String>>> recordsGroupsByNode = new ArrayList<>();

        for (int i = 0; i < subGeneralizationNodes.size(); i++) {
            recordsGroupsByNode.add(new ArrayList<>());
        }

        if (recordsGroupsByNode.isEmpty()) {
            // split is not necessary
            return new ArrayList<>();
        }

        for (List<String> records : partition.getRecords()) {
            String qIdValue = records.get(columnQId);

            boolean found = false;
            for (int i = 0; i < subGeneralizationNodes.size() && !found; i++) {
                GeneralizationNode node = subGeneralizationNodes.get(i);

                if (node.covers(qIdValue)) {
                    recordsGroupsByNode.get(i).add(records);
                    found = true;
                }
            }

            if (subGeneralizationNodes.isEmpty()) {
                System.out.println("Generalization hierarchy error!");
                throw new RuntimeException("Generalization hierarchy error!");
            }
        }

        boolean flag = true;

        for (int i = 0; i < recordsGroupsByNode.size() && flag; i++) {
            List<List<String>> recordsCoveredByNode = recordsGroupsByNode.get(i);

            if (recordsCoveredByNode.size() < k) {
                flag = false;
            }
        }

        if (flag) {
            for (int i = 0; i < recordsGroupsByNode.size(); i++) {
                List<List<String>> recordsCoveredByNode = recordsGroupsByNode.get(i);

                if (recordsCoveredByNode.size() > 0) {
                    List<Tuple<Integer, Integer>> wtemp = new ArrayList<>(pwidth);
                    CurrentGeneralizations newGeneralizations = currentGeneralizations.clone();

                    GeneralizationNode node = subGeneralizationNodes.get(i);
                    int numberOfLeafsCovered = node.getNumberOfLeafsCovered();
                    String value = node.getLabel();

                    wtemp.set(columnQId, new Tuple<>(numberOfLeafsCovered, numberOfLeafsCovered));
                    newGeneralizations.setGeneralization(columnQId, value);

                    subPartitions.add(new Partition(recordsCoveredByNode, wtemp, newGeneralizations, numberOfQIds));
                }
            }
        }

        return subPartitions;
    }


    public List<Partition> splitPartition(Partition partition, int columnQId) {
        /*
        split partition and distribute records to different sub-partitions
        */
        CurrentGeneralizations currentQIdsGeneralizations = partition.getCurrentQIdsGeneralization();

        if (!isGeneralizationCategorical.get(columnQId)) {
            List<Tuple<Integer, Integer>> pwidth = new ArrayList<>(partition.getWidth());
            return split_numerical(partition, columnQId, pwidth, currentQIdsGeneralizations);
        } else {
            List<Tuple<Integer, Integer>> pwidth = new ArrayList<>(partition.getWidth());

            return splitCategorical(partition, columnQId, pwidth, currentQIdsGeneralizations);
        }
    }

    public void anonymize(Partition partition) {
        /*
        Main procedure of Half_Partition.
        recursively partition groups until not allowable.
        */
        // print len(partition)
        // print partition.allow
        // pdb.set_trace()
        if (!isSplittable(partition)) {
            partitionsFinished.add(partition);
            return;
        }

        // Choose column of the QID
        int columnQId = chooseQId(partition);

        if (columnQId == -1) {
            System.out.println("Error: columnQId=-1");
            throw new RuntimeException("Error: columnQId=-1");
        }

        List<Partition> subPartitions = splitPartition(partition, columnQId);

        if (subPartitions.isEmpty()) {
            partition.disallow(columnQId);

            anonymize(partition);
        } else {
            for (Partition sub_p : subPartitions) {
                anonymize(sub_p);
            }
        }
    }

    public boolean isSplittable(Partition partition) {
    /*
    Check if the partition can be further split while satisfying k-anonymity.
    */
        return partition.anyAllowed();
    }

    public void init(List<Generalization> generalizations, int QI_num) {
        this.generalizations = generalizations;

        for (Generalization generalization : generalizations) {
            if (generalization.isRange()) {
                isGeneralizationCategorical.add(false);
            } else {
                isGeneralizationCategorical.add(true);
            }
        }

        numberOfQIds = QI_num;
        partitionsFinished = new ArrayList<>();
        QI_RANGE = new ArrayList<>();
    }

    public void execute(List<Generalization> generalizations, int QI_num) {
        init(generalizations, QI_num);

        CurrentGeneralizations overallGeneralization = new CurrentGeneralizations(nrTotalElements);
        List<List<String>> result = new ArrayList<>();
        List<Tuple<Integer, Integer>> wtemp = new ArrayList<>();

        for (int i = 0; i < numberOfQIds; i++) {
            Generalization generalization = this.generalizations.get(i);

            if (!isGeneralizationCategorical.get(i)) {
                GeneralizationRange numRange = (GeneralizationRange) generalization;
                QI_RANGE.add(numRange.getRange());
                wtemp.add(new Tuple<>(0, numRange.getMaxValue()));
                overallGeneralization.add(generalization);
            } else {
                int numberOfLeafsCovered = generalization.getNumberOfLeafsCovered();

                QI_RANGE.add(numberOfLeafsCovered);
                wtemp.add(new Tuple<>(numberOfLeafsCovered, numberOfLeafsCovered));
                overallGeneralization.add(generalization);
            }
        }

        Partition initialPartitionWithAll = new Partition(data, wtemp, overallGeneralization, numberOfQIds);
        anonymize(initialPartitionWithAll);

        double informationLossPercentage = 0;

        for (Partition partition : partitionsFinished) {

            transformIds(partition);

            for (int i = 0; i < partition.size(); i++) {
                List<String> notQIds = partition.getNotQIds(i);

                List<String> generalizationResult = new ArrayList<>(partition.getCurrentQIdsGeneralizationValues());
                generalizationResult.addAll(notQIds);

                result.add(generalizationResult);
            }
            informationLossPercentage = partition.getInformationLossPercentage();
        }

        printInformationLoss(informationLossPercentage);

        this.dataTransformed = result;
    }

    @Override
    public void addData(TableData data) {
        this.tableData = data;
        this.data = new ArrayList<>();

        for (int i = 0; i < data.getNumberOfRows(); i++) {
            Row row = data.getCurrentRow(i);

            ArrayList<String> rowData = row.getColumns();

            this.data.add(rowData);

            nrTotalElements += rowData.size();
        }
    }

    private void transformIds(Partition partition) {
        List<Integer> idPositions = tableData.getIdsPos();

        idPositions.forEach(partition::totallyAnonymize);
    }

    @Override
    public void apply() {
        int numberOfQIds = this.tableData.getQIdsNumber();

        execute(this.tableData.getGeneralizationsForColumns(), numberOfQIds);
    }

    @Override
    public TableData getDataTransformed() {
        TableData tableData = new TableData(this.tableData);

        for (List<String> rowData : dataTransformed) {
            Row row = new Row();

            for (String columnData : rowData) {
                row.addValue(columnData);
            }

            tableData.addRaw(row);
        }

        return tableData;
    }

    private void printInformationLoss(double informationLossPercentage) {
        System.out.println();
        System.out.printf("# %s%n", tableData.getTableName());
        System.out.println("-----------");
        System.out.printf("%3.1f%% information lost%n", informationLossPercentage * 100);
        System.out.println("-----------");
    }
}
